import numbers
import sympy as sp
import numpy as np

class Coordinates:
    """Class for handling curvilinear coordinates

    Parameters
    ----------
    psi : tuple
        The new coordinates
    rv : tuple
        The position vector in terms of the new coordinates
    """
    def __init__(self, psi, rv, assumptions=True):
        self._psi = psi
        self._rv = rv
        self._assumptions = assumptions
        self._hi = None
        self._b = None
        self._bt = None
        self._g = None
        self._gt = None
        self._ct = None
        self._det_g = {True: None, False: None}
        self._sqrt_det_g = {True: None, False: None}

    @property
    def b(self):
        return self.get_covariant_basis()

    @property
    def hi(self):
        return self.get_scaling_factors()

    @property
    def sg(self):
        return self.get_sqrt_det_g(True)

    @property
    def coordinates(self):
        return (self.psi, self.rv, self._assumptions)

    @property
    def psi(self):
        return self._psi

    @property
    def rv(self):
        return self._rv

    @property
    def is_orthogonal(self):
        return sp.Matrix(self.get_covariant_metric_tensor()).is_diagonal()

    @property
    def is_cartesian(self):
        return sp.Matrix(self.get_covariant_metric_tensor()).is_Identity

    def get_det_g(self, covariant=True):
        """Return determinant of covariant metric tensor"""
        if self._det_g[covariant] is not None:
            return self._det_g[covariant]
        if covariant:
            g = sp.Matrix(self.get_covariant_metric_tensor()).det()
        else:
            g = sp.Matrix(self.get_contravariant_metric_tensor()).det()
        g = sp.refine(g, self._assumptions)
        g = sp.simplify(g)
        self._det_g[covariant] = g
        return g

    def get_sqrt_det_g(self, covariant=True):
        """Return square root of determinant of covariant metric tensor"""
        if self._sqrt_det_g[covariant] is not None:
            return self._sqrt_det_g[covariant]
        g = self.get_det_g(covariant)
        sg = sp.refine(sp.simplify(sp.sqrt(g)), self._assumptions)
        if isinstance(sg, numbers.Number):
            if isinstance(sg, numbers.Real):
                sg = np.float(sg)
            elif isinstance(sg, numbers.Complex):
                sg = np.complex(sg)
            else:
                raise NotImplementedError
        self._sqrt_det_g[covariant] = sg
        return sg

    def get_cartesian_basis(self):
        """Return Cartesian basis vectors"""
        return np.eye(len(self.rv), dtype=object)

    def get_scaling_factors(self):
        """Return scaling factors"""
        if self._hi is not None:
            return self._hi
        hi = np.zeros_like(self.psi)
        u = sp.Wild('x')

        for i, s in enumerate(np.sum(self.b**2, axis=1)):
            hi[i] = sp.refine(sp.sqrt(sp.simplify(s)), self._assumptions)
            hi[i] = sp.simplify(hi[i].replace(sp.cosh(u)**2, 1+sp.sinh(u)**2))
            hi[i] = sp.simplify(hi[i].replace(sp.sinh(u)**2, (sp.cosh(2*u)-1)/2))

        self._hi = hi
        return hi

    def get_covariant_basis(self):
        """Return covariant basisvectors"""
        if self._b is not None:
            return self._b
        b = np.zeros((len(self.psi), len(self.rv)), dtype=object)
        for i, ti in enumerate(self.psi):
            for j, rj in enumerate(self.rv):
                b[i, j] = sp.simplify(rj.diff(ti, 1))
        self._b = b
        return b

    def get_contravariant_basis(self):
        """Return contravariant basisvectors"""
        if self._bt is not None:
            return self._bt
        bt = np.zeros_like(self.b)
        g = self.get_contravariant_metric_tensor()
        b = self.b
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                bt[i] += g[i, j]*b[j]

        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                bt[i, j] = sp.simplify(bt[i, j])
        self._bt = bt
        return bt

    def get_covariant_metric_tensor(self):
        """Return covariant metric tensor"""
        if self._g is not None:
            return self._g
        g = np.zeros((len(self.psi), len(self.psi)), dtype=object)
        b = self.b
        u = sp.Wild('x')
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                g[i, j] = sp.simplify(np.dot(b[i], b[j]))
                g[i, j] = sp.simplify(g[i, j].replace(sp.cosh(u)**2, 1+sp.sinh(u)**2))
                g[i, j] = sp.simplify(g[i, j].replace(sp.sinh(u)**2, (sp.cosh(2*u)-1)/2))
        self._g = g
        return g

    def get_contravariant_metric_tensor(self):
        """Return contravariant metric tensor"""
        if self._gt is not None:
            return self._gt
        g = self.get_covariant_metric_tensor()
        gt = sp.Matrix(g).inv()
        for i in range(gt.shape[0]):
            for j in range(gt.shape[1]):
                gt[i, j] = sp.simplify(gt[i, j])
        gt = np.array(gt)
        self._gt = gt
        return gt

    def get_christoffel_second(self):
        """Return Christoffel symbol of second kind"""
        if self._ct is not None:
            return self._ct
        b = self.get_covariant_basis()
        bt = self.get_contravariant_basis()
        ct = np.zeros((len(self.psi),)*3, object)
        for i in range(len(self.psi)):
            for j in range(len(self.psi)):
                for k in range(len(self.psi)):
                    ct[k, i, j] = sp.simplify(np.dot(np.array([bij.diff(self.psi[j], 1) for bij in b[i]]), bt[k]))
        self._ct = ct
        return ct

    def latex_basis_vectors(self, symbol_names=None, covariant=True):
        if covariant:
            b = self.get_covariant_basis()
        else:
            b = self.get_contravariant_basis()
        psi = self.psi
        symbols = {p: str(p) for p in psi}
        if symbol_names is not None:
            symbols = symbol_names

        k = {0: '\\mathbf{i}', 1: '\\mathbf{j}', 2: '\\mathbf{k}'}
        m = ' '
        for i, p in enumerate(psi):
            if covariant:
                m += '\\mathbf{b}_{%s} ='%(symbols[p])
            else:
                m += '\\mathbf{b}^{%s} ='%(symbols[p])
            for j in range(b.shape[1]):
                if b[i, j] == 1:
                    m += (k[j]+'+')
                elif b[i, j] != 0:
                    sl = sp.latex(b[i, j], symbol_names=symbols)
                    if sl.startswith('-') and not isinstance(b[i, j], sp.Add):
                        m = m.rstrip('+')
                    if isinstance(b[i, j], sp.Add):
                        sl = '\\left(%s\\right)'%(sl)
                    m += (sl+'\\,'+k[j]+'+')

            m = m.rstrip('+')
            m += ' \\\\ '
        m += ' '
        return r'%s'%(m)



import numba as nb

__all__ = ['TDMA_SymLU', 'TDMA_SymSolve', 'TDMA_SymSolve2D',
           'TDMA_SymSolve3D', 'TDMA_SymLU_VC', 'TDMA_SymSolve_VC',
           'TDMA_O_SymLU', 'TDMA_O_SymSolve']

#@nb.jit((float[:], float[:], float[:]), cache=True, nopython=True, fastmath=True)
@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymLU(d, ud, ld):
    n = d.shape[0]
    for i in range(2, n):
        ld[i-2] = ud[i-2]/d[i-2]
        d[i] -= ld[i-2]*ud[i-2]

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_O_SymLU(d, ud, ld):
    n = d.shape[0]
    for i in range(1, n):
        ld[i-1] = ud[i-1]/d[i-1]
        d[i] = d[i] - ld[i-1]*ud[i-1]

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymLU_2D(d, ud, ld, axis):
    if axis == 0:
        for j in range(d.shape[1]):
            TDMA_SymLU(d[:, j], ud[:, j], ld[:, j])
    elif axis == 1:
        for i in range(d.shape[0]):
            TDMA_SymLU(d[i], ud[i], ld[i])

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymLU_3D(d, ud, ld, axis):
    if axis == 0:
        for j in range(d.shape[1]):
            for k in range(d.shape[2]):
                TDMA_SymLU(d[:, j, k], ud[:, j, k], ld[:, j, k])
    elif axis == 1:
        for i in range(d.shape[0]):
            for k in range(d.shape[2]):
                TDMA_SymLU(d[i, :, k], ud[i, :, k], ld[i, :, k])
    elif axis == 2:
        for i in range(d.shape[0]):
            for j in range(d.shape[1]):
                TDMA_SymLU(d[i, j], ud[i, j], ld[i, j])

def TDMA_SymLU_VC(d, a, l, axis=0):
    n = d.ndim
    if n == 1:
        TDMA_SymLU(d, a, l)
    elif n == 2:
        TDMA_SymLU_2D(d, a, l, axis)
    elif n == 3:
        TDMA_SymLU_3D(d, a, l, axis)

def TDMA_SymSolve(d, a, l, x, axis=0):
    n = x.ndim
    if n == 1:
        TDMA_SymSolve1D(d, a, l, x)
    elif n == 2:
        TDMA_SymSolve2D(d, a, l, x, axis)
    elif n == 3:
        TDMA_SymSolve3D(d, a, l, x, axis)

def TDMA_SymSolve_VC(d, a, l, x, axis=0):
    n = x.ndim
    if n == 1:
        TDMA_SymSolve1D(d, a, l, x)
    elif n == 2:
        TDMA_SymSolve_VC_2D(d, a, l, x, axis)
    elif n == 3:
        TDMA_SymSolve_VC_3D(d, a, l, x, axis)

#@nb.jit((float[::1], float[::1], float[::1], float[::1]), nopython=True, fastmath=True)
@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymSolve1D(d, a, l, x):
    n = x.shape[0]-2
    for i in range(2, n):
        x[i] -= l[i-2]*x[i-2]

    x[n-1] = x[n-1]/d[n-1]
    x[n-2] = x[n-2]/d[n-2]
    for i in range(n - 3, -1, -1):
        x[i] = (x[i] - a[i]*x[i+2])/d[i]

#@nb.jit((float[:], float[:], float[:], complex[:, :], nb.int64), cache=True, nopython=True, fastmath=True)
@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymSolve2D(d, a, l, x, axis):
    n = d.shape[0]
    if axis == 0:
        for i in range(2, n):
            for j in range(x.shape[1]):
                x[i, j] -= l[i-2]*x[i-2, j]

        for j in range(x.shape[1]):
            x[n-1, j] = x[n-1, j]/d[n-1]
            x[n-2, j] = x[n-2, j]/d[n-2]

        for i in range(n - 3, -1, -1):
            d1 = 1./d[i]
            for j in range(x.shape[1]):
                x[i, j] = (x[i, j] - a[i]*x[i+2, j])*d1

    elif axis == 1:
        for i in range(x.shape[0]):
            TDMA_SymSolve1D(d, a, l, x[i])

#@nb.jit([(nb.float32[:], nb.float32[:], nb.float32[:], nb.complex64[:, :, :], nb.int64),
#         (nb.float64[:], nb.float64[:], nb.float64[:], nb.complex128[:, :, :], nb.int64)], cache=True, nopython=True, fastmath=True)
@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymSolve3D(d, a, l, x, axis):
    n = d.shape[0]
    if axis == 0:
        for i in range(2, n):
            for j in range(x.shape[1]):
                for k in range(x.shape[2]):
                    x[i, j, k] -= l[i-2]*x[i-2, j, k]

        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                x[n-1, j, k] = x[n-1, j, k]/d[n-1]
                x[n-2, j, k] = x[n-2, j, k]/d[n-2]

        for i in range(n - 3, -1, -1):
            d1 = 1./d[i]
            for j in range(x.shape[1]):
                for k in range(x.shape[2]):
                    x[i, j, k] = (x[i, j, k] - a[i]*x[i+2, j, k])*d1

    elif axis == 1:
        for i in range(x.shape[0]):
            for j in range(2, n):
                for k in range(x.shape[2]):
                    x[i, j, k] -= l[j-2]*x[i, j-2, k]

            for k in range(x.shape[2]):
                x[i, n-1, k] = x[i, n-1, k]/d[n-1]
                x[i, n-2, k] = x[i, n-2, k]/d[n-2]

            for j in range(n - 3, -1, -1):
                for k in range(x.shape[2]):
                    x[i, j, k] = (x[i, j, k] - a[j]*x[i, j+2, k])/d[j]

    elif axis == 2:
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                TDMA_SymSolve1D(d, a, l, x[i, j])

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymSolve_VC_3D(d, a, l, x, axis):
    if axis == 0:
        for j in range(d.shape[1]):
            for k in range(d.shape[2]):
                TDMA_SymSolve1D(d[:, j, k], a[:, j, k], l[:, j, k], x[:, j, k])
    elif axis == 1:
        for i in range(d.shape[0]):
            for k in range(d.shape[2]):
                TDMA_SymSolve1D(d[i, :, k], a[i, :, k], l[i, :, k], x[i, :, k])
    elif axis == 2:
        for i in range(d.shape[0]):
            for j in range(d.shape[1]):
                TDMA_SymSolve1D(d[i, j], a[i, j], l[i, j], x[i, j])

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_SymSolve_VC_2D(d, a, l, x, axis):
    if axis == 0:
        for j in range(d.shape[1]):
            TDMA_SymSolve1D(d[:, j], a[:, j], l[:, j], x[:, j])
    elif axis == 1:
        for i in range(d.shape[0]):
            TDMA_SymSolve1D(d[i], a[i], l[i], x[i])

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_O_SymSolve1D(d, a, l, x, axis=0):
    n = d.shape[0]
    for i in range(1, n):
        x[i] -= l[i-1]*x[i-1]
    x[n-1] = x[n-1]/d[n-1]
    for i in range(n - 2, -1, -1):
        x[i] = (x[i] - a[i]*x[i+1])/d[i]

@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_O_SymSolve2D(d, a, l, x, axis):
    n = d.shape[0]
    if axis == 0:
        for i in range(1, n):
            for j in range(x.shape[1]):
                x[i, j] -= l[i-1]*x[i-1, j]

        for j in range(x.shape[1]):
            x[n-1, j] = x[n-1, j]/d[n-1]

        for i in range(n - 2, -1, -1):
            d1 = 1./d[i]
            for j in range(x.shape[1]):
                x[i, j] = (x[i, j] - a[i]*x[i+1, j])*d1

    elif axis == 1:
        for i in range(x.shape[0]):
            TDMA_O_SymSolve1D(d, a, l, x[i])

#@nb.jit([(nb.float32[:], nb.float32[:], nb.float32[:], nb.complex64[:, :, :], nb.int64),
#         (nb.float64[:], nb.float64[:], nb.float64[:], nb.complex128[:, :, :], nb.int64)], cache=True, nopython=True, fastmath=True)
@nb.jit(nopython=True, fastmath=True, cache=True)
def TDMA_O_SymSolve3D(d, a, l, x, axis):
    n = d.shape[0]
    if axis == 0:
        for i in range(1, n):
            for j in range(x.shape[1]):
                for k in range(x.shape[2]):
                    x[i, j, k] -= l[i-1]*x[i-1, j, k]

        for j in range(x.shape[1]):
            for k in range(x.shape[2]):
                x[n-1, j, k] = x[n-1, j, k]/d[n-1]

        for i in range(n - 2, -1, -1):
            d1 = 1./d[i]
            for j in range(x.shape[1]):
                for k in range(x.shape[2]):
                    x[i, j, k] = (x[i, j, k] - a[i]*x[i+1, j, k])*d1

    elif axis == 1:
        for i in range(x.shape[0]):
            for j in range(1, n):
                for k in range(x.shape[2]):
                    x[i, j, k] -= l[j-1]*x[i, j-1, k]

            for k in range(x.shape[2]):
                x[i, n-1, k] = x[i, n-1, k]/d[n-1]

            for j in range(n - 2, -1, -1):
                for k in range(x.shape[2]):
                    x[i, j, k] = (x[i, j, k] - a[j]*x[i, j+1, k])/d[j]

    elif axis == 2:
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                TDMA_O_SymSolve1D(d, a, l, x[i, j])

def TDMA_O_SymSolve(d, a, l, x, axis=0):
    n = x.ndim
    if n == 1:
        TDMA_O_SymSolve1D(d, a, l, x)
    elif n == 2:
        TDMA_O_SymSolve2D(d, a, l, x, axis)
    elif n == 3:
        TDMA_O_SymSolve3D(d, a, l, x, axis)



"""
Module for implementing helper functions.
"""
import types
from numbers import Number
try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
from collections import defaultdict
import numpy as np
import sympy as sp
from scipy.fftpack import dct
from shenfun.optimization import optimizer

__all__ = ['dx', 'clenshaw_curtis1D', 'CachedArrayDict',
           'outer', 'apply_mask', 'integrate_sympy', 'mayavi_show']


def dx(u):
    r"""Compute integral of u over domain

    .. math::

        \int_{\Omega} u dx

    Parameters
    ----------

        u : Array
            The Array to integrate

    """
    T = u.function_space()
    uc = u.copy()
    dim = len(u.shape)
    if dim == 1:
        w = T.points_and_weights(weighted=False)[1]
        return np.sum(uc*w).item()

    for ax in range(dim):
        uc = uc.redistribute(axis=ax)
        w = T.bases[ax].points_and_weights(weighted=False)[1]
        sl = [np.newaxis]*len(uc.shape)
        sl[ax] = slice(None)
        uu = np.sum(uc*w[tuple(sl)], axis=ax)
        sl = [slice(None)]*len(uc.shape)
        sl[ax] = np.newaxis
        uc[:] = uu[tuple(sl)]
    return uc.flat[0]

def clenshaw_curtis1D(u, quad="GC"):  # pragma: no cover
    """Clenshaw-Curtis integration in 1D"""
    assert u.ndim == 1
    N = u.shape[0]
    if quad == 'GL':
        w = np.arange(0, N, 1, dtype=float)
        w[2:] = 2./(1-w[2:]**2)
        w[0] = 1
        w[1::2] = 0
        ak = dct(u, 1)
        ak /= (N-1)
        return np.sqrt(np.sum(ak*w))

    assert quad == 'GC'
    d = np.zeros(N)
    k = 2*(1 + np.arange((N-1)//2))
    d[::2] = (2./N)/np.hstack((1., 1.-k*k))
    w = dct(d, type=3)
    return np.sqrt(np.sum(u*w))

class CachedArrayDict(MutableMapping):
    """Dictionary for caching Numpy arrays (work arrays)

    Example
    -------

    >>> import numpy as np
    >>> from shenfun.utilities import CachedArrayDict
    >>> work = CachedArrayDict()
    >>> a = np.ones((3, 4), dtype=int)
    >>> w = work[(a, 0, True)] # create work array with shape as a
    >>> print(w.shape)
    (3, 4)
    >>> print(w)
    [[0 0 0 0]
     [0 0 0 0]
     [0 0 0 0]]
    >>> w2 = work[(a, 1, True)] # Get different(note 1!) array of same shape/dtype
    """
    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        newkey, fill = self.__keytransform__(key)
        try:
            value = self._data[newkey]
        except KeyError:
            shape, dtype, _ = newkey
            value = np.zeros(shape, dtype=np.dtype(dtype, align=True))
            self._data[newkey] = value
        if fill:
            value.fill(0)
        return value

    @staticmethod
    def __keytransform__(key):
        assert len(key) == 3
        return (key[0].shape, key[0].dtype, key[1]), key[2]

    def __len__(self):
        return len(self._data)

    def __setitem__(self, key, value):
        self._data[self.__keytransform__(key)[0]] = value

    def __delitem__(self, key):
        del self._data[self.__keytransform__(key)[0]]

    def __iter__(self):
        return iter(self._data)

    def values(self):
        raise TypeError('Cached work arrays not iterable')

def outer(a, b, c):
    r"""Return outer product $c_{i,j} = a_i b_j$

    Parameters
    ----------
    a : Array of shape (N, ...)
    b : Array of shape (N, ...)
    c : Array of shape (N*N, ...)

    The outer product is taken over the first index of a and b,
    for all remaining indices.
    """
    av = a.v
    bv = b.v
    cv = c.v
    symmetric = a is b
    if av.shape[0] == 2:
        outer2D(av, bv, cv, symmetric)
    elif av.shape[0] == 3:
        outer3D(av, bv, cv, symmetric)
    return c

@optimizer
def outer2D(a, b, c, symmetric):
    c[0] = a[0]*b[0]
    c[1] = a[0]*b[1]
    if symmetric:
        c[2] = c[1]
    else:
        c[2] = a[1]*b[0]
    c[3] = a[1]*b[1]

@optimizer
def outer3D(a, b, c, symmetric):
    c[0] = a[0]*b[0]
    c[1] = a[0]*b[1]
    c[2] = a[0]*b[2]
    if symmetric:
        c[3] = c[1]
        c[6] = c[2]
        c[7] = c[5]
    else:
        c[3] = a[1]*b[0]
        c[6] = a[2]*b[0]
        c[7] = a[2]*b[1]
    c[4] = a[1]*b[1]
    c[5] = a[1]*b[2]
    c[8] = a[2]*b[2]

@optimizer
def apply_mask(u_hat, mask):
    if mask is not None:
        u_hat *= mask
    return u_hat

def integrate_sympy(f, d):
    """Exact definite integral using sympy

    Try to convert expression `f` to a polynomial before integrating.

    See sympy issue https://github.com/sympy/sympy/pull/18613 to why this is
    needed. Poly().integrate() is much faster than sympy.integrate() when applicable.

    Parameters
    ----------
    f : sympy expression
    d : 3-tuple
        First item the symbol, next two the lower and upper integration limits
    """
    try:
        p = sp.Poly(f, d[0]).integrate()
        return p(d[2]) - p(d[1])
    except sp.PolynomialError:
        #return sp.Integral(f, d).evalf()
        return sp.integrate(f, d)

def split(measures):
    #ms = sp.sympify(measures).expand()
    #ms = ms if isinstance(ms, tuple) else [ms]
    #result = []
    #for m in ms:
    #    if sp.simplify(m) == 0:
    #        continue
    #    d = {'coeff': m} if isinstance(m, Number) else sp.separatevars(m, dict=True)
    #    d = defaultdict(lambda: 1, {str(k): sp.simplify(v) for k, v in d.items()})
    #    dc = d['coeff']
    #    d['coeff'] = int(dc) if isinstance(dc, (sp.Integer, int)) else float(dc)
    #    result.append(d)
    #return result

    def _split(mss, result):
        for ms in mss:
            ms = sp.sympify(ms)
            if isinstance(ms, sp.Mul):
                # Multiplication of two or more terms
                result = _split(ms.args, result)
                continue

            # Something else with only one symbol
            sym = ms.free_symbols
            assert len(sym) <= 1
            if len(sym) == 1:
                sym = sym.pop()
                result[str(sym)] *= ms
            else:
                ms = int(ms) if isinstance(ms, sp.Integer) else float(ms)
                result['coeff'] *= ms
        return result

    ms = sp.sympify(measures).expand()
    result = []
    if isinstance(ms, sp.Add):
        for arg in ms.args:
            result.append(_split([arg], defaultdict(lambda: 1)))
    else:
        result.append(_split([ms], defaultdict(lambda: 1)))
    return result

def mayavi_show():
    """
    Return show function that updates the mayavi figure in the background.
    """
    from pyface.api import GUI
    from mayavi import mlab
    return mlab.show(GUI().stop_event_loop)



"""
This module contains the implementation of operators acting on arguments.
"""
import numpy as np
import sympy as sp
import copy
from .arguments import Expr, BasisFunction

__all__ = ('div', 'grad', 'Dx', 'curl')

#pylint: disable=protected-access

def _expr_from_vector_components(comp, basis):
    """Return Expr composed of vector components `comp`
    """
    terms, scales, indices = [], [], []
    for i in range(len(comp)):
        terms += comp[i]._terms
        scales += comp[i]._scales
        indices += comp[i]._indices
    return Expr(basis, terms, scales, indices)

def div(test):
    """Return div(test)

    Parameters
    ----------
    test:  Expr or BasisFunction
           Must be rank > 0 (cannot take divergence of scalar)

    """
    assert isinstance(test, (Expr, BasisFunction))

    if isinstance(test, BasisFunction):
        test = Expr(test)

    ndim = test.dimensions
    coors = test.function_space().coors

    if coors.is_cartesian:

        if ndim == 1:      # 1D
            v = np.array(test.terms())
            v += 1
            test._terms = v.tolist()
            return test

        else:
            if test.num_components() == ndim**2: # second rank tensor
                dv = []
                for i in range(ndim):
                    dv.append(div(test[i]))

                return _expr_from_vector_components(dv, test.basis())

            else: # vector
                d = Dx(test[0], 0, 1)
                for i in range(1, ndim):
                    d += Dx(test[i], i, 1)
                d.simplify()
                return d

    else:

        if ndim == 1:      # 1D
            sg = coors.get_sqrt_det_g()
            d = Dx(test*sg, 0, 1)*(1/sg)
            return d

        else:
            if test.num_components() == ndim**2:

                ct = coors.get_christoffel_second()
                d = []
                for i in range(ndim):
                    di = []
                    for j in range(ndim):
                        Sij = test[i][j]
                        di.append(Dx(Sij, j, 1))
                        for k in range(ndim):
                            Skj = test[k][j]
                            if not ct[i, j, k] == 0:
                                di.append(Skj*ct[i, j, k])
                            if not ct[k, k, j] == 0:
                                di.append(Sij*ct[k, k, j])

                    dj = di[0]
                    for j in range(1, len(di)):
                        dj += di[j]
                    dj.simplify()
                    d.append(dj)
                return _expr_from_vector_components(d, test.basis())

            else:
                sg = coors.get_sqrt_det_g()
                d = Dx(test[0]*sg, 0, 1)*(1/sg)
                for i in range(1, ndim):
                    d += Dx(test[i]*sg, i, 1)*(1/sg)
            d.simplify()
        return d

def grad(test):
    """Return grad(test)

    Parameters
    ----------
    test: Expr or BasisFunction

    Note
    ----
    Increases the rank of Expr by one

    """
    assert isinstance(test, (Expr, BasisFunction))

    if isinstance(test, BasisFunction):
        test = Expr(test)

    ndim = test.dimensions
    coors = test.function_space().coors

    if coors.is_cartesian:

        d = []
        if test.num_components() > 1:
            for i in range(test.num_components()):
                for j in range(ndim):
                    d.append(Dx(test[i], j, 1))
        else:
           for i in range(ndim):
               d.append(Dx(test, i, 1))

    else:

        gt = coors.get_contravariant_metric_tensor()

        if test.num_components() > 1:
            ct = coors.get_christoffel_second()
            d = []
            for i in range(ndim):
                vi = test[i]
                for j in range(ndim):
                    dj = []
                    for l in range(ndim):
                        sc = gt[l, j]
                        if not sc == 0:
                            dj.append(Dx(vi, l, 1)*sc)
                        for k in range(ndim):
                            if not sc*ct[i, k, l] == 0:
                                dj.append(test[k]*(sc*ct[i, k, l]))

                    di = dj[0]
                    for j in range(1, len(dj)):
                        di += dj[j]
                    d.append(di)

        else:
            d = []
            for i in range(ndim):
                dj = []
                for j in range(ndim):
                    sc = gt[j, i]
                    if not sc == 0:
                        dj.append(Dx(test, j, 1)*sc)
                di = dj[0]
                for j in range(1, len(dj)):
                    di += dj[j]
                d.append(di)

    dv = _expr_from_vector_components(d, test.basis())
    dv.simplify()
    return dv

def Dx(test, x, k=1):
    """Return k'th order partial derivative in direction x

    Parameters
    ----------
    test: Expr or BasisFunction
    x:  int
        axis to take derivative over
    k:  int
        Number of derivatives
    """
    assert isinstance(test, (Expr, BasisFunction))

    if k > 1:
        for l in range(k):
            test = Dx(test, x, 1)
        return test

    if isinstance(test, BasisFunction):
        test = Expr(test)

    test = copy.copy(test)
    coors = test.function_space().coors

    if coors.is_cartesian:
        v = np.array(test.terms())
        v[..., x] += k
        test._terms = v.tolist()

    else:
        assert test.expr_rank() < 1, 'Cannot (yet) take derivative of tensor in curvilinear coordinates'
        psi = coors.psi
        v = copy.deepcopy(test.terms())
        sc = copy.deepcopy(test.scales())
        ind = copy.deepcopy(test.indices())
        num_terms = test.num_terms()
        for i in range(test.num_components()):
            for j in range(num_terms[i]):
                sc0 = sp.simplify(sp.diff(sc[i][j], psi[x], k))
                sc0 = sp.refine(sc0, coors._assumptions)

                if not sc0 == 0:
                    v[i].append(copy.deepcopy(v[i][j]))
                    sc[i].append(sc0)
                    ind[i].append(ind[i][j])
                v[i][j][x] += k
        test._terms = v
        test._scales = sc
        test._indices = ind

    return test

def curl(test):
    """Return curl of test

    Parameters
    ----------
    test: Expr or BasisFunction

    """
    assert isinstance(test, (Expr, BasisFunction))

    if isinstance(test, BasisFunction):
        test = Expr(test)

    test = copy.copy(test)

    assert test.expr_rank() > 0
    assert test.num_components() == test.dimensions

    coors = test.function_space().coors

    if coors.is_cartesian:
        if test.dimensions == 3:
            w0 = Dx(test[2], 1, 1) - Dx(test[1], 2, 1)
            w1 = Dx(test[0], 2, 1) - Dx(test[2], 0, 1)
            w2 = Dx(test[1], 0, 1) - Dx(test[0], 1, 1)
            test._terms = w0.terms()+w1.terms()+w2.terms()
            test._scales = w0.scales()+w1.scales()+w2.scales()
            test._indices = w0.indices()+w1.indices()+w2.indices()
        else:
            assert test.dimensions == 2
            test = Dx(test[1], 0, 1) - Dx(test[0], 1, 1)

    else:
        assert test.expr_rank() < 2, 'Cannot (yet) take curl of higher order tensor in curvilinear coordinates'
        hi = coors.hi
        sg = coors.get_sqrt_det_g()
        if coors.is_orthogonal:
            if test.dimensions == 3:
                w0 = (Dx(test[2]*hi[2]**2, 1, 1) - Dx(test[1]*hi[1]**2, 2, 1))*(1/sg)
                w1 = (Dx(test[0]*hi[0]**2, 2, 1) - Dx(test[2]*hi[2]**2, 0, 1))*(1/sg)
                w2 = (Dx(test[1]*hi[1]**2, 0, 1) - Dx(test[0]*hi[0]**2, 1, 1))*(1/sg)
                test = _expr_from_vector_components([w0, w1, w2], test.basis())
            else:
                assert test.dimensions == 2
                test = (Dx(test[1]*hi[1]**2, 0, 1) - Dx(test[0]*hi[0]**2, 1, 1))*(1/sg)
        else:
            g = coors.get_covariant_metric_tensor()

            if test.dimensions == 3:
                ct = coors.get_christoffel_second()
                w0 = np.sum([(Dx(test[i]*g[2, i], 1, 1) - Dx(test[i]*g[1, i], 2, 1))*(1/sg) for i in range(3)])
                w1 = np.sum([(Dx(test[i]*g[0, i], 2, 1) - Dx(test[i]*g[2, i], 0, 1))*(1/sg) for i in range(3)])
                w2 = np.sum([(Dx(test[i]*g[1, i], 0, 1) - Dx(test[i]*g[0, i], 1, 1))*(1/sg) for i in range(3)])
                # Don't think this double loop is needed due to symmetry of Christoffel?
                #for i in range(3):
                #    for k in range(3):
                #        w0 += (ct[i, 1, 2]*g[i, k]*test[k] - ct[i, 2, 1]*g[i, k]*test[k])*(1/sg)
                #        w1 += (ct[i, 2, 0]*g[i, k]*test[k] - ct[i, 0, 2]*g[i, k]*test[k])*(1/sg)
                #        w2 += (ct[i, 0, 1]*g[i, k]*test[k] - ct[i, 1, 0]*g[i, k]*test[k])*(1/sg)

                # This is an alternative (more complicated way):
                #gt = coors.get_contravariant_metric_tensor()
                #ww0 = grad(g[0, 0]*test[0] + g[0, 1]*test[1] + g[0, 2]*test[2])
                #ww1 = grad(g[1, 0]*test[0] + g[1, 1]*test[1] + g[1, 2]*test[2])
                #ww2 = grad(g[2, 0]*test[0] + g[2, 1]*test[1] + g[2, 2]*test[2])
                #d0 = sg*(ww0[1]*gt[0, 2] + ww1[1]*gt[1, 2] + ww2[1]*gt[2, 2] - ww0[2]*gt[0, 1] - ww1[2]*gt[1, 1] - ww2[2]*gt[2, 1])
                #d1 = sg*(ww0[2]*gt[0, 0] + ww1[2]*gt[1, 0] + ww2[2]*gt[2, 0] - ww0[0]*gt[0, 2] - ww1[0]*gt[1, 2] - ww2[0]*gt[2, 2])
                #d2 = sg*(ww0[0]*gt[0, 1] + ww1[0]*gt[1, 1] + ww2[0]*gt[2, 1] - ww0[1]*gt[0, 0] - ww1[1]*gt[1, 0] - ww2[1]*gt[2, 0])
                #w0 = d0*gt[0, 0] + d1*gt[1, 0] + d2*gt[2, 0]
                #w1 = d0*gt[0, 1] + d1*gt[1, 1] + d2*gt[2, 1]
                #w2 = d0*gt[0, 2] + d1*gt[1, 2] + d2*gt[2, 2]

                test = _expr_from_vector_components([w0, w1, w2], test.basis())
            else:
                assert test.dimensions == 2
                test = np.sum([(Dx(test[i]*g[1, i], 0, 1) - Dx(test[i]*g[0, i], 1, 1))*(1/sg) for i in range(2)])

    test.simplify()
    return test



"""
Module for function spaces of generalized Jacobi type

Note the environment variable

    SHENFUN_JACOBI_MODE

that can be used for extended precision in this module.

SHENFUN_JACOBI_MODE = 'mpmath' will use the extended precision
mpmath module to compute inner products.

The precision can be set using, e.g.,

    from mpmath import mp
    mp.dps = 50

where mp.dps is the number of significant digits.

Note that extended precision is costly, but for some of the
matrices that can be created with the Jacobi bases it is necessary.
Also note the the higher precision is only used for assembling
matrices comuted with :func:`evaluate_basis_derivative_all`.
It has no effect for the matrices that are predifined in the
matrices.py module. Also note that the final matrix will be
in regular double precision. So the higher precision is only used
for the intermediate assembly.

"""

import os
import functools
import numpy as np
import sympy as sp
from scipy.special import eval_jacobi, roots_jacobi #, gamma
from mpi4py_fft import fftw
from shenfun.spectralbase import SpectralBase, Transform, islicedict, slicedict
from shenfun.forms.arguments import Function
from shenfun.chebyshev.bases import BCBiharmonic, BCDirichlet

try:
    import quadpy
    from mpmath import mp
    mp.dps = 30
    has_quadpy = True
except:
    has_quadpy = False
    mp = None

mode = os.environ.get('SHENFUN_JACOBI_MODE', 'numpy')
mode = mode if has_quadpy else 'numpy'

_x = sp.Symbol('x', real=True)

#pylint: disable=method-hidden,no-else-return,not-callable,abstract-method,no-member,cyclic-import

__all__ = ['JacobiBase', 'Orthogonal', 'ShenDirichlet', 'ShenBiharmonic',
           'ShenOrder6', 'mode', 'has_quadpy', 'mp']


class JacobiBase(SpectralBase):
    """Base class for all Jacobi spaces

    Parameters
    ----------
        N : int
            Number of quadrature points
        quad : str, optional
            Type of quadrature

            - JG - Jacobi-Gauss
        alpha : number, optional
            Parameter of the Jacobi polynomial
        beta : number, optional
            Parameter of the Jacobi polynomial
        domain : 2-tuple of floats, optional
            The computational domain
        padding_factor : float, optional
            Factor for padding backward transforms.
        dealias_direct : bool, optional
            Set upper 1/3 of coefficients to zero before backward transform
        dtype : data-type, optional
            Type of input data in real physical space. Will be overloaded when
            basis is part of a :class:`.TensorProductSpace`.
        coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
            Map for curvilinear coordinatesystem.
            The new coordinate variable in the new coordinate system is the first item.
            Second item is a tuple for the Cartesian position vector as function of the
            new variable in the first tuple. Example::

                theta = sp.Symbols('x', real=True, positive=True)
                rv = (sp.cos(theta), sp.sin(theta))

    """

    def __init__(self, N, quad="JG", alpha=0, beta=0, domain=(-1., 1.), dtype=np.float,
                 padding_factor=1, dealias_direct=False, coordinates=None):
        SpectralBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype,
                              padding_factor=padding_factor, dealias_direct=dealias_direct,
                              coordinates=coordinates)
        self.alpha = alpha
        self.beta = beta
        self.forward = functools.partial(self.forward, fast_transform=False)
        self.backward = functools.partial(self.backward, fast_transform=False)
        self.scalar_product = functools.partial(self.scalar_product, fast_transform=False)

    @staticmethod
    def family():
        return 'jacobi'

    def reference_domain(self):
        return (-1, 1)

    def get_refined(self, N):
        return self.__class__(N,
                              quad=self.quad,
                              domain=self.domain,
                              dtype=self.dtype,
                              padding_factor=self.padding_factor,
                              dealias_direct=self.dealias_direct,
                              coordinates=self.coors.coordinates,
                              alpha=self.alpha,
                              beta=self.beta)

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        return self.__class__(self.N,
                              quad=self.quad,
                              domain=self.domain,
                              dtype=self.dtype,
                              padding_factor=padding_factor,
                              dealias_direct=dealias_direct,
                              coordinates=self.coors.coordinates,
                              alpha=self.alpha,
                              beta=self.beta)

    def get_orthogonal(self):
        return Orthogonal(self.N,
                          quad=self.quad,
                          domain=self.domain,
                          dtype=self.dtype,
                          padding_factor=self.padding_factor,
                          dealias_direct=self.dealias_direct,
                          coordinates=self.coors.coordinates,
                          alpha=0,
                          beta=0)

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        points, weights = roots_jacobi(N, self.alpha, self.beta)
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def mpmath_points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if mode == 'numpy' or not has_quadpy:
            return self.points_and_weights(N=N, map_true_domain=map_true_domain, weighted=weighted, **kw)
        if N is None:
            N = self.shape(False)
        pw = quadpy.c1.gauss_jacobi(N, self.alpha, self.beta, 'mpmath')
        points = pw.points_symbolic
        weights = pw.weights_symbolic
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def jacobi(self, x, alpha, beta, N):
        V = np.zeros((x.shape[0], N))
        if mode == 'numpy':
            for n in range(N):
                V[:, n] = eval_jacobi(n, alpha, beta, x)
        else:
            for n in range(N):
                V[:, n] = sp.lambdify(_x, sp.jacobi(n, alpha, beta, _x), 'mpmath')(x)
        return V

    def derivative_jacobi(self, x, alpha, beta, k=1):
        V = self.jacobi(x, alpha+k, beta+k, self.N)
        if k > 0:
            Vc = np.zeros_like(V)
            for j in range(k, self.N):
                dj = np.prod(np.array([j+alpha+beta+1+i for i in range(k)]))
                #dj = gamma(j+alpha+beta+1+k) / gamma(j+alpha+beta+1)
                Vc[:, j] = (dj/2**k)*V[:, j-k]
            V = Vc
        return V

    def vandermonde(self, x):
        return self.jacobi(x, self.alpha, self.beta, self.shape(False))

    def plan(self, shape, axis, dtype, options):
        if shape in (0, (0,)):
            return

        if isinstance(axis, tuple):
            assert len(axis) == 1
            axis = axis[0]

        if isinstance(self.forward, Transform):
            if self.forward.input_array.shape == shape and self.axis == axis:
                # Already planned
                return

        U = fftw.aligned(shape, dtype=dtype)
        V = fftw.aligned(shape, dtype=dtype)
        U.fill(0)
        V.fill(0)
        self.axis = axis
        if self.padding_factor > 1.+1e-8:
            trunc_array = self._get_truncarray(shape, V.dtype)
            self.forward = Transform(self.forward, None, U, V, trunc_array)
            self.backward = Transform(self.backward, None, trunc_array, V, U)
        else:
            self.forward = Transform(self.forward, None, U, V, V)
            self.backward = Transform(self.backward, None, V, V, U)
        self.scalar_product = Transform(self.scalar_product, None, U, V, V)
        self.si = islicedict(axis=self.axis, dimensions=self.dimensions)
        self.sl = slicedict(axis=self.axis, dimensions=self.dimensions)


class Orthogonal(JacobiBase):
    """Function space for regular (orthogonal) Jacobi functions

    Parameters
    ----------
        N : int
            Number of quadrature points
        quad : str, optional
            Type of quadrature

            - JG - Jacobi-Gauss

        padding_factor : float, optional
            Factor for padding backward transforms.
        dealias_direct : bool, optional
            Set upper 1/3 of coefficients to zero before backward transform
        dtype : data-type, optional
            Type of input data in real physical space. Will be overloaded when
            basis is part of a :class:`.TensorProductSpace`.
        coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
            Map for curvilinear coordinatesystem.
            The new coordinate variable in the new coordinate system is the first item.
            Second item is a tuple for the Cartesian position vector as function of the
            new variable in the first tuple. Example::

                theta = sp.Symbols('x', real=True, positive=True)
                rv = (sp.cos(theta), sp.sin(theta))
    """

    def __init__(self, N, quad="JG", alpha=-0.5, beta=-0.5, domain=(-1., 1.),
                 dtype=np.float, padding_factor=1, dealias_direct=False, coordinates=None):
        JacobiBase.__init__(self, N, quad=quad, alpha=alpha, beta=beta, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @property
    def is_orthogonal(self):
        return True

    #def get_orthogonal(self):
    #    return self

    def sympy_basis(self, i=0, x=_x):
        return sp.jacobi(i, self.alpha, self.beta, x)

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        if mode == 'numpy':
            output_array = eval_jacobi(i, self.alpha, self.beta, x, out=output_array)
        else:
            f = self.sympy_basis(i, _x)
            output_array[:] = sp.lambdify(_x, f, 'mpmath')(x)
        return output_array

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.points_and_weights(mode=mode)[0]
        #x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)

        if mode == 'numpy':
            dj = np.prod(np.array([i+self.alpha+self.beta+1+j for j in range(k)]))
            output_array[:] = dj/2**k*eval_jacobi(i-k, self.alpha+k, self.beta+k, x)
        else:
            f = sp.jacobi(i, self.alpha, self.beta, _x)
            output_array[:] = sp.lambdify(_x, f.diff(_x, k), 'mpmath')(x)
        return output_array

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mpmath_points_and_weights(mode=mode)[0]
        #if x.dtype == 'O':
        #    x = np.array(x, dtype=self.dtype)
        if mode == 'numpy':
            return self.derivative_jacobi(x, self.alpha, self.beta, k)
        else:
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            for i in range(N):
                V[:, i] = self.evaluate_basis_derivative(x, i, k, output_array=V[:, i])
        return V

    def evaluate_basis_all(self, x=None, argument=0):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        return self.vandermonde(x)


class ShenDirichlet(JacobiBase):
    """Jacobi function space for Dirichlet boundary conditions

    Parameters
    ----------
        N : int
            Number of quadrature points
        quad : str, optional
            Type of quadrature

            - JG - Jacobi-Gauss

        bc : tuple of numbers
            Boundary conditions at edges of domain
        domain : 2-tuple of floats, optional
            The computational domain
        padding_factor : float, optional
            Factor for padding backward transforms.
        dealias_direct : bool, optional
            Set upper 1/3 of coefficients to zero before backward transform
        dtype : data-type, optional
            Type of input data in real physical space. Will be overloaded when
            basis is part of a :class:`.TensorProductSpace`.
        coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
            Map for curvilinear coordinatesystem.
            The new coordinate variable in the new coordinate system is the first item.
            Second item is a tuple for the Cartesian position vector as function of the
            new variable in the first tuple. Example::

                theta = sp.Symbols('x', real=True, positive=True)
                rv = (sp.cos(theta), sp.sin(theta))

    """
    def __init__(self, N, quad='JG', bc=(0, 0), domain=(-1., 1.), dtype=np.float,
                 padding_factor=1, dealias_direct=False, coordinates=None):
        JacobiBase.__init__(self, N, quad=quad, alpha=-1, beta=-1, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        assert bc in ((0, 0), 'Dirichlet')
        from shenfun.tensorproductspace import BoundaryValues
        self._bc_basis = None
        self.bc = BoundaryValues(self, bc=bc)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    def get_refined(self, N):
        return self.__class__(N,
                              quad=self.quad,
                              domain=self.domain,
                              dtype=self.dtype,
                              padding_factor=self.padding_factor,
                              dealias_direct=self.dealias_direct,
                              coordinates=self.coors.coordinates,
                              bc=self.bc.bc)

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        return self.__class__(self.N,
                              quad=self.quad,
                              domain=self.domain,
                              dtype=self.dtype,
                              padding_factor=padding_factor,
                              dealias_direct=dealias_direct,
                              coordinates=self.coors.coordinates,
                              bc=self.bc.bc)

    def is_scaled(self):
        return False

    def slice(self):
        return slice(0, self.N-2)

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        N = self.shape(False)
        V = np.zeros((x.shape[0], N))
        for i in range(N-2):
            V[:, i] = self.evaluate_basis_derivative(x, i, k, output_array=V[:, i])
        return V

    def sympy_basis(self, i=0, x=_x):
        return (1-x**2)*sp.jacobi(i, 1, 1, x)
        #return (1-x)**(-self.alpha)*(1+x)**(-self.beta)*sp.jacobi(i, -self.alpha, -self.beta, x)

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)
        f = self.sympy_basis(i, _x)
        output_array[:] = sp.lambdify(_x, f.diff(_x, k), mode)(x)
        return output_array

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)
        if mode == 'numpy':
            output_array = (1-x**2)*eval_jacobi(i, -self.alpha, -self.beta, x, out=output_array)
        else:
            f = self.sympy_basis(i, _x)
            output_array[:] = sp.lambdify(_x, f, 'mpmath')(x)
        return output_array

    def evaluate_basis_all(self, x=None, argument=0):
        if mode == 'numpy':
            if x is None:
                x = self.mesh(False, False)
            V = np.zeros((x.shape[0], self.N))
            V[:, :-2] = self.jacobi(x, 1, 1, self.N-2)*(1-x**2)[:, np.newaxis]
        else:
            if x is None:
                x = self.mpmath_points_and_weights()[0]
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            for i in range(N-2):
                V[:, i] = self.evaluate_basis(x, i, output_array=V[:, i])
        return V

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        points, weights = roots_jacobi(N, self.alpha+1, self.beta+1)
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def mpmath_points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if mode == 'numpy' or not has_quadpy:
            return self.points_and_weights(N=N, map_true_domain=map_true_domain, weighted=weighted, **kw)
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        pw = quadpy.c1.gauss_jacobi(N, self.alpha+1, self.beta+1, mode)
        points = pw.points_symbolic
        weights = pw.weights_symbolic
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def to_ortho(self, input_array, output_array=None):
        assert self.alpha == -1 and self.beta == -1
        if output_array is None:
            output_array = Function(input_array.function_space().get_orthogonal())
        else:
            output_array.fill(0)
        k = self.wavenumbers().astype(np.float)
        s0 = self.sl[slice(0, -2)]
        s1 = self.sl[slice(2, None)]
        z = input_array[s0]*2*(k+1)/(2*k+3)
        output_array[s0] = z
        output_array[s1] -= z
        return output_array

    def _evaluate_scalar_product(self, fast_transform=True):
        SpectralBase._evaluate_scalar_product(self)
        self.scalar_product.output_array[self.sl[slice(-2, None)]] = 0

    def get_bc_basis(self):
        if self._bc_basis:
            return self._bc_basis
        self._bc_basis = BCDirichlet(self.N, quad=self.quad, domain=self.domain,
                                     coordinates=self.coors.coordinates)
        return self._bc_basis


class ShenBiharmonic(JacobiBase):
    """Function space for Biharmonic boundary conditions

    Parameters
    ----------
        N : int
            Number of quadrature points
        quad : str, optional
            Type of quadrature

            - JG - Jacobi-Gauss

        domain : 2-tuple of floats, optional
            The computational domain
        padding_factor : float, optional
            Factor for padding backward transforms.
        dealias_direct : bool, optional
            Set upper 1/3 of coefficients to zero before backward transform
        dtype : data-type, optional
            Type of input data in real physical space. Will be overloaded when
            basis is part of a :class:`.TensorProductSpace`.
        coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
            Map for curvilinear coordinatesystem.
            The new coordinate variable in the new coordinate system is the first item.
            Second item is a tuple for the Cartesian position vector as function of the
            new variable in the first tuple. Example::

                theta = sp.Symbols('x', real=True, positive=True)
                rv = (sp.cos(theta), sp.sin(theta))

    Note
    ----
    The generalized Jacobi function j^{alpha=-2, beta=-2} is used as basis. However,
    inner products are computed without weights, for alpha=beta=0.

    """
    def __init__(self, N, quad='JG', bc=(0, 0, 0, 0), domain=(-1., 1.), dtype=np.float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        JacobiBase.__init__(self, N, quad=quad, alpha=-2, beta=-2, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        assert bc in ((0, 0, 0, 0), 'Biharmonic')
        from shenfun.tensorproductspace import BoundaryValues
        self._bc_basis = None
        self.bc = BoundaryValues(self, bc=bc)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def boundary_condition():
        return 'Biharmonic'

    def slice(self):
        return slice(0, self.N-4)

    def sympy_basis(self, i=0, x=_x):
        return (1-x**2)**2*sp.jacobi(i, 2, 2, x)

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        N = self.shape(False)
        V = np.zeros((x.shape[0], N))
        for i in range(N-4):
            V[:, i] = self.evaluate_basis_derivative(x, i, k, output_array=V[:, i])
        return V

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)
        f = self.sympy_basis(i, _x)
        output_array[:] = sp.lambdify(_x, f.diff(_x, k), mode)(x)
        return output_array

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        if mode == 'numpy':
            output_array[:] = (1-x**2)**2*eval_jacobi(i, 2, 2, x, out=output_array)
        else:
            X = sp.symbols('x', real=True)
            f = self.sympy_basis(i, X)
            output_array[:] = sp.lambdify(X, f, 'mpmath')(x)
        return output_array

    def evaluate_basis_all(self, x=None, argument=0):
        if mode == 'numpy':
            if x is None:
                x = self.mesh(False, False)
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            V[:, :-4] = self.jacobi(x, 2, 2, N-4)*((1-x**2)**2)[:, np.newaxis]
        else:
            if x is None:
                x = self.mpmath_points_and_weights()[0]
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            for i in range(N-4):
                V[:, i] = self.evaluate_basis(x, i, output_array=V[:, i])
        return V

    def _evaluate_scalar_product(self, fast_transform=True):
        SpectralBase._evaluate_scalar_product(self)
        self.scalar_product.output_array[self.sl[slice(-4, None)]] = 0

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        points, weights = roots_jacobi(N, 0, 0)
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def mpmath_points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if mode == 'numpy' or not has_quadpy:
            return self.points_and_weights(N=N, map_true_domain=map_true_domain, weighted=weighted, **kw)
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        pw = quadpy.c1.gauss_jacobi(N, 0, 0, 'mpmath')
        points = pw.points_symbolic
        weights = pw.weights_symbolic
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def to_ortho(self, input_array, output_array=None):
        if output_array is None:
            output_array = Function(input_array.function_space().get_orthogonal())
        else:
            output_array.fill(0)
        k = self.wavenumbers().astype(np.float)
        _factor0 = 4*(k+2)*(k+1)/(2*k+5)/(2*k+3)
        _factor1 = (-2*(2*k+5)/(2*k+7))
        _factor2 = ((2*k+3)/(2*k+7))
        s0 = self.sl[slice(0, -4)]
        z = _factor0*input_array[s0]
        output_array[s0] = z
        output_array[self.sl[slice(2, -2)]] += z*_factor1
        output_array[self.sl[slice(4, None)]] += z*_factor2
        return output_array

    def get_bc_basis(self):
        if self._bc_basis:
            return self._bc_basis
        self._bc_basis = BCBiharmonic(self.N, quad=self.quad, domain=self.domain,
                                      coordinates=self.coors.coordinates)
        return self._bc_basis


class ShenOrder6(JacobiBase):
    """Function space for 6th order equation

    Parameters
    ----------
        N : int
            Number of quadrature points
        quad : str, optional
            Type of quadrature

            - JG - Jacobi-Gauss

        domain : 2-tuple of floats, optional
            The computational domain
        padding_factor : float, optional
            Factor for padding backward transforms.
        dealias_direct : bool, optional
            Set upper 1/3 of coefficients to zero before backward transform
        dtype : data-type, optional
            Type of input data in real physical space. Will be overloaded when
            basis is part of a :class:`.TensorProductSpace`.
        coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
            Map for curvilinear coordinatesystem.
            The new coordinate variable in the new coordinate system is the first item.
            Second item is a tuple for the Cartesian position vector as function of the
            new variable in the first tuple. Example::

                theta = sp.Symbols('x', real=True, positive=True)
                rv = (sp.cos(theta), sp.sin(theta))

    Note
    ----
    The generalized Jacobi function j^{alpha=-3, beta=-3} is used as basis. However,
    inner products are computed without weights, for alpha=beta=0.

    """
    def __init__(self, N, quad='JG', domain=(-1., 1.), dtype=np.float, padding_factor=1, dealias_direct=False,
                 coordinates=None):
        JacobiBase.__init__(self, N, quad=quad, alpha=-3, beta=-3, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        from shenfun.tensorproductspace import BoundaryValues
        self.bc = BoundaryValues(self, bc=(0,)*6)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def boundary_condition():
        return '6th order'

    def slice(self):
        return slice(0, self.N-6)

    def sympy_basis(self, i=0, x=_x):
        return (1-x**2)**3*sp.jacobi(i, 3, 3, x)

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        if output_array is None:
            output_array = np.zeros(x.shape)
        f = self.sympy_basis(i, _x)
        output_array[:] = sp.lambdify(_x, f.diff(_x, k), mode)(x)
        return output_array

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mpmath_points_and_weights()[0]
        N = self.shape(False)
        V = np.zeros((x.shape[0], N))
        for i in range(N-6):
            V[:, i] = self.evaluate_basis_derivative(x, i, k, output_array=V[:, i])
        return V

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        if mode == 'numpy':
            output_array[:] = (1-x**2)**3*eval_jacobi(i, 3, 3, x, out=output_array)
        else:
            f = self.sympy_basis(i, _x)
            output_array[:] = sp.lambdify(_x, f, 'mpmath')(x)
        return output_array

    def evaluate_basis_all(self, x=None, argument=0):
        if mode == 'numpy':
            if x is None:
                x = self.mesh(False, False)
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            V[:, :-6] = self.jacobi(x, 3, 3, N-6)*((1-x**2)**3)[:, np.newaxis]
        else:
            if x is None:
                x = self.mpmath_points_and_weights()[0]
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            for i in range(N-6):
                V[:, i] = self.evaluate_basis(x, i, output_array=V[:, i])
        return V

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        points, weights = roots_jacobi(N, 0, 0)
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def mpmath_points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if mode == 'numpy' or not has_quadpy:
            return self.points_and_weights(N=N, map_true_domain=map_true_domain, weighted=weighted, **kw)
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        pw = quadpy.c1.gauss_jacobi(N, 0, 0, 'mpmath')
        points = pw.points_symbolic
        weights = pw.weights_symbolic
        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def get_orthogonal(self):
        return Orthogonal(self.N, alpha=0, beta=0, dtype=self.dtype, domain=self.domain, coordinates=self.coors.coordinates)

    def _evaluate_scalar_product(self, fast_transform=True):
        SpectralBase._evaluate_scalar_product(self)
        self.scalar_product.output_array[self.sl[slice(-6, None)]] = 0

    #def to_ortho(self, input_array, output_array=None):
    #    if output_array is None:
    #        output_array = np.zeros_like(input_array.v)
    #    k = self.wavenumbers().astype(np.float)
    #    _factor0 = 4*(k+2)*(k+1)/(2*k+5)/(2*k+3)
    #    _factor1 = (-2*(2*k+5)/(2*k+7))
    #    _factor2 = ((2*k+3)/(2*k+7))
    #    s0 = self.sl[slice(0, -4)]
    #    z = _factor0*input_array[s0]
    #    output_array[s0] = z
    #    output_array[self.sl[slice(2, -2)]] -= z*_factor1
    #    output_array[self.sl[slice(4, None)]] += z*_factor2
    #    return output_array
